// ==================================================================
//           Author: Jean-Michel RICHER
//            Email: jean-michel.richer@univ-angers.fr
//      Institution: LERIA, Computer Science Department
//                   Faculty of Sciences, 
//                   University of Angers 
//                   France 
// Last Modified on: December 2021
// ==================================================================
// PROGRAM
//
//   Display Julia set.
//
// ==================================================================

#include <iostream>
#include <sstream>
#include <vector>
#include <algorithm>
#include <stdexcept>
using namespace std;
#include <getopt.h>
#include "gl_animation.h"
#include "cpu_timer.h"
#include "complex.h"
#include "tiny_lib.h"

// //////////////////////////////////////////////////////////////////
// GLOBAL VARIABLES
// //////////////////////////////////////////////////////////////////

// width of the image
int IMAGE_WIDTH = 1024;
// height of the image
int IMAGE_HEIGHT = 1024;
// display image or perform test if false
bool view_flag = false;
// maximum number of images to compute
int max_images = 200;
// maximum number of terms to compute in the
// series @formula( z_{n+1}^2 = z_n^2 + z_0 )
int max_terms = 400;
// real part of the constant @formula( z_0 )
REAL constant_real = 0.3;
// imaginary part of the constant @formula( z_0 )
REAL constant_imag = 0.5;

// reference of the images
REAL xmin = -1.4;
REAL xmax = 1.4;
REAL ymin = -1.5;
REAL ymax = 1.5;

// limit over which the series does not converge
REAL NORM_THRESOLD = 4.0;

// name of the last image saved on disk
string output_file_name = "image";


GLImage *image;


/* ==================================================================
 * SUBPROGRAM
 *
 *  REAL julia( int x, int y );
 *
 * WHAT
 * 
 *  Determine for a given point of the image if the formula
 *  @formula( z_{n+1} = z_n^2 + z_0 ) converges ot not.
 *
 * PARAMETERS
 *
 *  @paramdef(int , x) : abscissa of the point
 *  @paramdef(int , y) : ordinate of the point
 *
 * ==================================================================
 */ 
REAL julia( int x, int y ) {

	// transpose the point in the image into the reference
	// of the complex terms
	REAL r = x * (REAL)(xmax - xmin)/IMAGE_WIDTH + xmin;
	REAL i = y * (REAL)(ymax - ymin)/IMAGE_HEIGHT + ymin;

	aComplex z(r, i);
	aComplex c(constant_real, constant_imag);

	// compute series until maximum number of terms has been reached
	// of term is greater than threshold
	for (int k = 1; k <= max_terms; k++) {
	
		z = z * z + c;
	
		if (z.norm() > NORM_THRESOLD)  {
	
			return NORM_THRESOLD;
	
		}
	
	}

	return z.norm();
	
}

/* ==================================================================
 * SUBPROGRAM
 *
 *  void kernel( GLImage *image );
 *
 * WHAT
 * 
 *  Subprogram that computes the color of each pixel of the image
 *  depending on the convergence of the series 
 *  @formula( z_{n+1} = z_n^2 + z_0 ).
 *
 * PARAMETERS
 *
 *  @paramdef(GLImage *, image) : image to fill
 *
 * ==================================================================
 */ 
void kernel( GLImage *image ) {

	ImagePoint *pixels = image->pixels();

	// for each pixel find if series converges
	for (int y=0; y<IMAGE_HEIGHT; y++) {
	
		for (int x=0; x<IMAGE_WIDTH; x++) {

			int offset = (x + y * IMAGE_WIDTH);
			
			REAL juliaValue = julia( x, y );

			int  color = round( juliaValue * 255 );
			color = 255 - color;
			
			pixels[ offset + 0 ].red =  color;
			pixels[ offset + 1 ].green =  color;
			pixels[ offset + 2 ].blue = color;
			pixels[ offset + 3 ].alpha = 255;
		}
	}
	
}

/* ==================================================================
 * SUBPROGRAM
 *
 *  void animate();
 *
 * WHAT
 * 
 *  Subprogram called if we use the default mode of the application.
 *  The default mode consists in displaying the resulting image while
 *  the test mode does not display the image.
 *
 * ==================================================================
 */ 
void animate() {

	cout << "frame=" << GLAnimation::instance().frames() << endl;
	double factor = 1.001;
	
	xmax *= factor;
	ymax *= factor;
	xmin *= factor;
	ymin *= factor;
	
	kernel( image );
	
	if (GLAnimation::instance().frames() == static_cast<size_t>( max_images )) {

		cout << "frames=" << max_images << endl;
		double animation_time = GLAnimation::instance().time();
		cout << "timer=" <<  animation_time << endl;
		double fps = max_images / animation_time;
		cout << "frames_per_seconds=" << fps << endl;
	
		if (output_file_name.length() > 0) {
			
#ifdef LIBPNGPP
			image->save_png( output_file_name + ".png");
#else		
			image->save( output_file_name + ".ppm");	
#endif
		}
		
		exit( EXIT_SUCCESS );
	}
	

}

/* ==================================================================
 * SUBPROGRAM
 *
 *  void leave(); 
 *
 * WHAT
 * 
 *  Subprogram called when the animation ends
 *
 * ==================================================================
 */ 
 void leave() {
	
	cout << "frames.end=" << GLAnimation::instance().frames() << endl;
	cout << "timer=" << GLAnimation::instance().time() << endl;

}


/* ==================================================================
 * SUBPROGRAM
 *
 *  void help( char *argv[] );
 *
 * WHAT
 * 
 *  Display help about command line arguments
 *
 * ==================================================================
 */
void help(char *argv[]) {

	string nl2 = "\n\n";
	string bu = "\33[4m"; // begin underline mode
	string eu = "\033[0m"; // end underline mode
	string bb = "\033[1m"; // begin bold mode
	string eb = "\033[0m"; // end bold mode
	
	cout << nl2 << bb << "NAME" << eb << "\n\t" << argv[0] << nl2;
	cout << bb << "SYNOPSIS" << eb << "\n\t" << argv[0] << " [OPTION]...\n\n";
	cout << bb << "DESCRIPTION" << eb << "\n\n\t";
	cout << "Display spheres in a 3D environement using a simple ray-tracing approach\n\n";
	
	help_clopt_wa( 'w', "width", "integer", 
		"set width of image computed between 128 and 8192" );
	help_clopt_wa( 'y', "height", "integer", 
		"set height of image computed between 128 and 8192" );
	
	help_clopt_wa( 'a', "max-images", "integer", 
		"set number times the image is calculated between 100 and 10000" );
	help_clopt_wa( 't', "max-terms", "integer", 
		"set number of terms calculated to evaluate convergence between 10 and 1000" );
	
	help_clopt_wa( 'r', "constant-real", "float", "set real part of constant z0" );
	help_clopt_wa( 'i', "constant-imag", "float", "set imaginary part of constant z0" );
	
	help_clopt_na( 'v', "view", "display image otherwise test" );
	help_clopt_na( 'h', "help", "print help about this program" );
	
	exit( EXIT_FAILURE );
	
}


/* ==================================================================
 * SUBPROGRAM
 *
 *  int main( int argc, char *argv[] )
 *
 * WHAT
 * 
 *  Main program
 *
 * PARAMETERS
 *
 *  @paramdef(int,argc) number of command line arguments
 *  @paramdef(char **,argv) array of strings that represent command 
 *	             line arguments
 *
 * ==================================================================
 */
int main( int argc, char *argv[] ) {
	
	// --------------------------------------------------------------
	// program command line arguments
	// --------------------------------------------------------------
	static struct option long_options[] = {
		{"width"		, required_argument, 0, 'w'},
		{"height"		, required_argument, 0, 'y'},
		{"view"			, 		no_argument, 0, 'v'},
		{"max-images"	, required_argument, 0, 'a'},
		{"max-terms"	, required_argument, 0, 't'},
		{"constant-real", required_argument, 0, 'r'},
		{"constant-imag", required_argument, 0, 'i'},
		{"help"			, 		no_argument, 0, 'h'},
		{0,0,0,0}
	}; 

	while (true) {
		int option_index = 0;
		int c = getopt_long(argc, argv, "a:w:y:i:r:s:vh", long_options, &option_index);
		if (c == -1) break;
		
		switch(c) {

		case 'w':
			set_from_range( IMAGE_WIDTH, atoi(optarg), 128, 8192, "image width" );
			break;
		
		case 'y':
			set_from_range( IMAGE_HEIGHT, atoi(optarg), 128, 8192, "image height" );
			break;

		case 'v':
			view_flag = true;
			break;

		case 'a':
			set_from_range( max_images, atoi(optarg), 10, 10000, "number of images" );
			break;
			
		case 't':
			set_from_range( max_terms, atoi(optarg), 10, 1000, "number of terms" );
			break;
			

		case 'r':
			constant_real = atof(optarg);
			break;

		case 'i':
			constant_imag = atof(optarg);
			break;
			
		case 'h':
			help( argv );
			break;
				
		default:
			help( argv );	
			break;
		}
	}

	cout << "image_width=" << IMAGE_WIDTH << endl;
	cout << "image_height=" << IMAGE_HEIGHT << endl;
	cout << "max_images=" << max_images << endl;
	cout << "max_terms=" << max_terms << endl;
	cout << "constant_real=" << constant_real << endl;
	cout << "constant_imag=" << constant_imag << endl;

	image = new GLImage( IMAGE_WIDTH, IMAGE_HEIGHT );
		
	// display image or perform test 
	if (view_flag) {	
	
		kernel( image );	
		GLAnimation::instance().image(image);	
		GLAnimation::instance().run(argc, argv, animate,  leave);
		
	} else {
	
		CPUTimer computation_timer;
		
		computation_timer.start();
	
		for (int i = 0; i < max_images; ++i) {
			kernel( image );
		}

		computation_timer.stop();
		cout << "timer.computation.ms=" << computation_timer.ms() << endl;
				
	}

	return 0;
	
}

